/*
 *  linux/drivers/mmc/samcop_sdi.c - Samsung SAMCOP SDI Interface driver
 *
 *  Copyright (C) 2004 Thomas Kleffel, All Rights Reserved.
 *  Copyright (C) 2005 Matt Reimer, All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <linux/blkdev.h>
#include <linux/delay.h>
#include <linux/err.h>
#include <linux/dma-mapping.h>
#include <linux/mmc/host.h>
#include <linux/mmc/protocol.h>
#include <linux/soc-device.h>

#include <asm/dma.h>
#include <asm/dma-mapping.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/hardware/amba.h>
#include <asm/hardware/clock.h>
#include <asm/mach/mmc.h>

#include <asm/hardware/ipaq-samcop.h>
#include <asm/hardware/ipaq-hamcop.h>
#include <asm/hardware/samcop-dma.h>
#include <asm/hardware/hamcop-dma.h>
#include <asm/hardware/samcop-sdi.h>
#include <asm/hardware/hamcop-sdi.h>
#include <asm/hardware/shamcop-clocks.h>
#include "../../arch/arm/common/ipaq/hamcop_base.h"
#include "../../arch/arm/common/ipaq/samcop_base.h"

#include <asm/arch/pxa-regs.h>
#include <asm/arch/h5400-asic.h>
#include <asm/arch/h5400-gpio.h>
#include <asm/arch-pxa/h2200-asic.h>
#include <asm/arch-pxa/h2200-gpio.h>

#include "samcop_sdi.h"

#ifdef CONFIG_MMC_DEBUG
#define DBG(x...)       printk(KERN_DEBUG x)
#else
#define DBG(x...)       do { } while (0)
#endif

#define DRIVER_NAME "samcop-sdi"
#define PFX DRIVER_NAME ": "

#define RESSIZE(ressource) (((ressource)->end - (ressource)->start)+1)

// #define KERN_DEBUG KERN_INFO


static struct samcop_dma_client samcop_sdi_dma_client = {
	.name		= DRIVER_NAME " DMA",
};


static u32
sdi_rdreg(struct samcop_sdi_host *host, u32 reg)
{

	if (host->asic_type == SAMCOP_SDI_ASIC_SAMCOP)
		return readl(host->base + reg);
	else {

		switch (reg) {

		case SAMCOP_SDICON:
			return __raw_readw(host->base + HAMCOP_SDICON);
			break;

		case SAMCOP_SDIPRE:
			return __raw_readw(host->base + HAMCOP_SDIPRE);
			break;

		case SAMCOP_SDICMDARG:
			return (__raw_readw(host->base + HAMCOP_SDICMDARGH) << 16) |
				__raw_readw(host->base + HAMCOP_SDICMDARGL);
			break;

		case SAMCOP_SDICMDCON:
			return __raw_readw(host->base + HAMCOP_SDICMDCON);
			break;

		case SAMCOP_SDICMDSTAT:
			return __raw_readw(host->base + HAMCOP_SDICMDSTAT);
			break;

		case SAMCOP_SDIRSP0:
			return (__raw_readw(host->base + HAMCOP_SDIRSP0H) << 16) |
				__raw_readw(host->base + HAMCOP_SDIRSP0L);
			break;

		case SAMCOP_SDIRSP1:
			return (__raw_readw(host->base + HAMCOP_SDIRSP1H) << 16) |
				__raw_readw(host->base + HAMCOP_SDIRSP1L);
			break;

		case SAMCOP_SDIRSP2:
			return (__raw_readw(host->base + HAMCOP_SDIRSP2H) << 16) |
				__raw_readw(host->base + HAMCOP_SDIRSP2L);
			break;

		case SAMCOP_SDIRSP3:
			return (__raw_readw(host->base + HAMCOP_SDIRSP3H) << 16) |
				__raw_readw(host->base + HAMCOP_SDIRSP3L);
			break;

		case SAMCOP_SDITIMER:
			return (__raw_readw(host->base + HAMCOP_SDITIMER1) << 16) |
				__raw_readw(host->base + HAMCOP_SDITIMER2);
			break;

		case SAMCOP_SDIBSIZE:
			return __raw_readw(host->base + HAMCOP_SDIBSIZE);
			break;

		case SAMCOP_SDIDCON:
			return (__raw_readw(host->base + HAMCOP_SDIDCON) << 12) |
			       (__raw_readw(host->base + HAMCOP_SDIBNUM) & 0xfff);
			break;

		case SAMCOP_SDIDCNT:
			return (__raw_readw(host->base + HAMCOP_SDIDCNT1) << 12) |
			       (__raw_readw(host->base + HAMCOP_SDIDCNT2) & 0xfff);
			break;

		case SAMCOP_SDIDSTA:
			return __raw_readw(host->base + HAMCOP_SDIDSTA);
			break;

		case SAMCOP_SDIFSTA:
			return __raw_readw(host->base + HAMCOP_SDIFSTA);
			break;

		case SAMCOP_SDIDATA:
			return (__raw_readw(host->base + HAMCOP_SDIDATA) << 16) |
				__raw_readw(host->base + HAMCOP_SDIDATA);
			break;

		case SAMCOP_SDIIMSK:
			return (__raw_readw(host->base + HAMCOP_SDIIMSK1) << 5) |
			       (__raw_readw(host->base + HAMCOP_SDIIMSK2) & 0x1f);
			break;
		}
	}

	return 0;
}


static void
sdi_wrreg(struct samcop_sdi_host *host, u32 reg, u32 val)
{
	if (host->asic_type == SAMCOP_SDI_ASIC_SAMCOP)
		writel(val, host->base + reg);
	else {

		switch (reg) {

		case SAMCOP_SDICON:
			__raw_writew(val, host->base + HAMCOP_SDICON);
			break;

		case SAMCOP_SDIPRE:
			__raw_writew(val, host->base + HAMCOP_SDIPRE);
			break;

		case SAMCOP_SDICMDARG:
			__raw_writew(val >> 16,    host->base + HAMCOP_SDICMDARGH);
			__raw_writew(val & 0xffff, host->base + HAMCOP_SDICMDARGL);
			break;

		case SAMCOP_SDICMDCON:
			__raw_writew(val, host->base + HAMCOP_SDICMDCON);
			break;

		case SAMCOP_SDICMDSTAT:
			__raw_writew(val, host->base + HAMCOP_SDICMDSTAT);
			break;

		case SAMCOP_SDIRSP0:
			__raw_writew(val >> 16,    host->base + HAMCOP_SDIRSP0H);
			__raw_writew(val & 0xffff, host->base + HAMCOP_SDIRSP0L);
			break;

		case SAMCOP_SDIRSP1:
			__raw_writew(val >> 16,    host->base + HAMCOP_SDIRSP1H);
			__raw_writew(val & 0xffff, host->base + HAMCOP_SDIRSP1L);
			break;

		case SAMCOP_SDIRSP2:
			__raw_writew(val >> 16,    host->base + HAMCOP_SDIRSP2H);
			__raw_writew(val & 0xffff, host->base + HAMCOP_SDIRSP2L);
			break;

		case SAMCOP_SDIRSP3:
			__raw_writew(val >> 16,    host->base + HAMCOP_SDIRSP3H);
			__raw_writew(val & 0xffff, host->base + HAMCOP_SDIRSP3L);
			break;

		case SAMCOP_SDITIMER:
			__raw_writew(val >> 16,    host->base + HAMCOP_SDITIMER1);
			__raw_writew(val & 0xffff, host->base + HAMCOP_SDITIMER2);
			break;

		case SAMCOP_SDIBSIZE:
			__raw_writew(val, host->base + HAMCOP_SDIBSIZE);
			break;

		case SAMCOP_SDIDCON:
			__raw_writew(val & 0xfff, host->base + HAMCOP_SDIBNUM);
			__raw_writew(val >> 12,   host->base + HAMCOP_SDIDCON);
			break;

		case SAMCOP_SDIDCNT:
			__raw_writew(val >> 12,   host->base + HAMCOP_SDIDCNT1);
			__raw_writew(val & 0xfff, host->base + HAMCOP_SDIDCNT2);
			break;

		case SAMCOP_SDIDSTA:
			__raw_writew(val, host->base + HAMCOP_SDIDSTA);
			break;

		case SAMCOP_SDIFSTA:
			__raw_writew(val, host->base + HAMCOP_SDIFSTA);
			break;

		case SAMCOP_SDIDATA:
			__raw_writew(val & 0xffff, host->base + HAMCOP_SDIDATA);
			__raw_writew(val >> 16,    host->base + HAMCOP_SDIDATA);
			break;

		case SAMCOP_SDIIMSK:
			__raw_writew(val >> 5,   host->base + HAMCOP_SDIIMSK1);
			__raw_writew(val & 0x1f, host->base + HAMCOP_SDIIMSK2);
			break;
		}
	}

}


/*
 * ISR for SDI Interface IRQ
 * Communication between driver and ISR works as follows:
 *   host->mrq 			points to current request
 *   host->complete_what	tells the ISR when the request is considered done
 *     COMPLETION_CMDSENT	  when the command was sent
 *     COMPLETION_RSPFIN          when a response was received
 *     COMPLETION_XFERFINISH	  when the data transfer is finished
 *     COMPLETION_XFERFINISH_RSPFIN both of the above.
 *   host->complete_request	is the completion-object the driver waits for
 *
 * 1) Driver sets up host->mrq and host->complete_what
 * 2) Driver prepares the transfer
 * 3) Driver enables interrupts
 * 4) Driver starts transfer
 * 5) Driver waits for host->complete_rquest
 * 6) ISR checks for request status (errors and success)
 * 6) ISR sets host->mrq->cmd->error and host->mrq->data->error
 * 7) ISR completes host->complete_request
 * 8) ISR disables interrupts
 * 9) Driver wakes up and takes care of the request
*/

static irqreturn_t samcop_sdi_irq(int irq, void *dev_id, struct pt_regs *regs)
{
	struct samcop_sdi_host *host;
	u32 sdi_csta, sdi_dsta, sdi_dcnt, sdi_fsta;
	u32 sdi_cclear, sdi_dclear;
	unsigned long iflags;

	host = (struct samcop_sdi_host *)dev_id;

	//Check for things not supposed to happen
	if(!host) return IRQ_HANDLED;
	
	sdi_csta 	= sdi_rdreg(host, SAMCOP_SDICMDSTAT);
	sdi_dsta 	= sdi_rdreg(host, SAMCOP_SDIDSTA);
	sdi_dcnt 	= sdi_rdreg(host, SAMCOP_SDIDCNT);
	sdi_fsta 	= sdi_rdreg(host, SAMCOP_SDIFSTA);
	
	DBG(PFX "IRQ csta=0x%08x dsta=0x%08x dcnt:0x%08x fsta:0x%08x\n", sdi_csta, sdi_dsta, sdi_dcnt, sdi_fsta);
		
	spin_lock_irqsave( &host->complete_lock, iflags);
	
	if( host->complete_what==COMPLETION_NONE ) {
		goto clear_imask;
	}
	
	if(!host->mrq) { 
		goto clear_imask;
	}

	
	sdi_csta 	= sdi_rdreg(host, SAMCOP_SDICMDSTAT);
	sdi_dsta 	= sdi_rdreg(host, SAMCOP_SDIDSTA);
	sdi_dcnt 	= sdi_rdreg(host, SAMCOP_SDIDCNT);
	sdi_cclear	= 0;
	sdi_dclear	= 0;
	
	
	if(sdi_csta & SAMCOP_SDICMDSTAT_CMDTIMEOUT) {
		host->mrq->cmd->error = MMC_ERR_TIMEOUT;
		goto transfer_closed;
	}

	if(sdi_csta & SAMCOP_SDICMDSTAT_CMDSENT) {
		if(host->complete_what == COMPLETION_CMDSENT) {
			host->mrq->cmd->error = MMC_ERR_NONE;
			goto transfer_closed;
		}

		sdi_cclear |= SAMCOP_SDICMDSTAT_CMDSENT;
	}

	if(sdi_csta & SAMCOP_SDICMDSTAT_CRCFAIL) {

		/* S3C* cores improperly detect CRC errors for commands with
		   long responses (CMD2, 9, 10), so ignore them. */
		if(host->mrq->cmd->flags & MMC_RSP_CRC &&
		   !host->mrq->cmd->flags & MMC_RSP_LONG) {
			host->mrq->cmd->error = MMC_ERR_BADCRC;
			goto transfer_closed;
		}

		sdi_cclear |= SAMCOP_SDICMDSTAT_CRCFAIL;
	}

	if(sdi_csta & SAMCOP_SDICMDSTAT_RSPFIN) {
		if(host->complete_what == COMPLETION_RSPFIN) {
			host->mrq->cmd->error = MMC_ERR_NONE;
			goto transfer_closed;
		}

		if(host->complete_what == COMPLETION_XFERFINISH_RSPFIN) {
			host->mrq->cmd->error = MMC_ERR_NONE;
			host->complete_what = COMPLETION_XFERFINISH;
		}

		sdi_cclear |= SAMCOP_SDICMDSTAT_RSPFIN;
	}

	if(sdi_dsta & SAMCOP_SDIDSTA_FIFOFAIL) {
		host->mrq->cmd->error = MMC_ERR_NONE;
		host->mrq->data->error = MMC_ERR_FIFO;
		goto transfer_closed;
	}

	if(sdi_dsta & SAMCOP_SDIDSTA_RXCRCFAIL) {
		host->mrq->cmd->error = MMC_ERR_NONE;
		host->mrq->data->error = MMC_ERR_BADCRC;
		goto transfer_closed;
	}

	if(sdi_dsta & SAMCOP_SDIDSTA_CRCFAIL) {
		host->mrq->cmd->error = MMC_ERR_NONE;
		host->mrq->data->error = MMC_ERR_BADCRC;
		goto transfer_closed;
	}

	if(sdi_dsta & SAMCOP_SDIDSTA_DATATIMEOUT) {
		host->mrq->cmd->error = MMC_ERR_NONE;
		host->mrq->data->error = MMC_ERR_TIMEOUT;
		goto transfer_closed;
	}

	if(sdi_dsta & SAMCOP_SDIDSTA_XFERFINISH) {
		if(host->complete_what == COMPLETION_XFERFINISH) {
			host->mrq->cmd->error = MMC_ERR_NONE;
			host->mrq->data->error = MMC_ERR_NONE;
			goto transfer_closed;
		}

		if(host->complete_what == COMPLETION_XFERFINISH_RSPFIN) {
			host->mrq->data->error = MMC_ERR_NONE;
			host->complete_what = COMPLETION_RSPFIN;
		}

		sdi_dclear |= SAMCOP_SDIDSTA_XFERFINISH;
	}

	sdi_wrreg(host, SAMCOP_SDICMDSTAT, sdi_cclear);
	sdi_wrreg(host, SAMCOP_SDIDSTA, sdi_dclear);

	spin_unlock_irqrestore( &host->complete_lock, iflags);
	DBG(PFX "IRQ still waiting.\n");
	return IRQ_HANDLED;


transfer_closed:
	host->complete_what = COMPLETION_NONE;
	complete(&host->complete_request);
	sdi_wrreg(host, SAMCOP_SDIIMSK, 0);
	spin_unlock_irqrestore( &host->complete_lock, iflags);
	DBG(PFX "IRQ transfer closed.\n");
	return IRQ_HANDLED;
	
clear_imask:
	sdi_wrreg(host, SAMCOP_SDIIMSK, 0);
	spin_unlock_irqrestore( &host->complete_lock, iflags);
	DBG(PFX "IRQ clear imask.\n");
	return IRQ_HANDLED;

}


/*
 * ISR for the CardDetect Pin
*/

static irqreturn_t samcop_sdi_irq_cd(int irq, void *dev_id, struct pt_regs *regs)
{
	struct samcop_sdi_host *host = (struct samcop_sdi_host *)dev_id;
	mmc_detect_change(host->mmc);

	return IRQ_HANDLED;
}



void samcop_sdi_dma_done_callback(samcop_dma_chan_t *dma_ch, void *buf_id,
	int size, samcop_dma_buffresult_t result)
{	unsigned long iflags;
	u32 sdi_csta, sdi_dsta,sdi_dcnt, sdi_fsta;
	struct samcop_sdi_host *host = (struct samcop_sdi_host *)buf_id;
	
	sdi_csta 	= sdi_rdreg(host, SAMCOP_SDICMDSTAT);
	sdi_dsta 	= sdi_rdreg(host, SAMCOP_SDIDSTA);
	sdi_dcnt 	= sdi_rdreg(host, SAMCOP_SDIDCNT);
	sdi_fsta 	= sdi_rdreg(host, SAMCOP_SDIFSTA);
	
	DBG(PFX "DMAD csta=0x%08x dsta=0x%08x dcnt:0x%08x fsta:0x%08x result:0x%08x\n", sdi_csta, sdi_dsta, sdi_dcnt, sdi_fsta, result);
	
	spin_lock_irqsave( &host->complete_lock, iflags);
	
	if(!host->mrq) goto out;
	if(!host->mrq->data) goto out;
	
	
	sdi_csta 	= sdi_rdreg(host, SAMCOP_SDICMDSTAT);
	sdi_dsta 	= sdi_rdreg(host, SAMCOP_SDIDSTA);
	sdi_dcnt 	= sdi_rdreg(host, SAMCOP_SDIDCNT);
		
	if( result!=SAMCOP_RES_OK ) {
		goto fail_request;
	}
	
	
	if(host->mrq->data->flags & MMC_DATA_READ) {
		if( sdi_dcnt>0 ) {
			goto fail_request;
		}
	}
	
out:	
	complete(&host->complete_dma);
	spin_unlock_irqrestore( &host->complete_lock, iflags);
	return;


fail_request:
	host->mrq->data->error = MMC_ERR_DMA;
	host->complete_what = COMPLETION_NONE;
	complete(&host->complete_dma);
	complete(&host->complete_request);
	sdi_wrreg(host, SAMCOP_SDIIMSK, 0);
	goto out;

}


void samcop_sdi_dma_setup(struct samcop_sdi_host *host, struct mmc_data *data)
{
	unsigned long devaddr;
	u32 hwsrcsel;
	int xfer_unit;

	if (host->asic_type == SAMCOP_SDI_ASIC_SAMCOP) {
		devaddr = host->mem->start + SAMCOP_SDIDATA;
		hwsrcsel = SAMCOP_DCON_CH0_SD;
		xfer_unit = 4;
	} else {
		devaddr = host->mem->start + HAMCOP_SDIDATA;
		hwsrcsel = SAMCOP_DCON_CH0_SD_HAMCOP;
		xfer_unit = 2;
	}

	samcop_dma_devconfig(host->dma, data->flags & MMC_DATA_READ ?
			     SAMCOP_DMASRC_HW : SAMCOP_DMASRC_MEM,
			     3, // XXX magic for APB | FIXED/NO_INC
			     devaddr);
	samcop_dma_config(host->dma, xfer_unit, SAMCOP_DCON_HWTRIG | hwsrcsel);
	samcop_dma_set_buffdone_fn(host->dma, samcop_sdi_dma_done_callback);
	samcop_dma_setflags(host->dma, SAMCOP_DMAF_AUTOSTART);
}

static void samcop_sdi_request(struct mmc_host *mmc, struct mmc_request *mrq) {
 	struct samcop_sdi_host *host = mmc_priv(mmc);
	u32 sdi_carg, sdi_ccon, sdi_timer;
	u32 sdi_bsize, sdi_dcon, sdi_imsk;
	int dma_len;

	DBG(KERN_DEBUG PFX "request: [CMD] opcode:0x%02x arg:0x%08x flags:%x retries:%u\n",
		mrq->cmd->opcode, mrq->cmd->arg, mrq->cmd->flags, mrq->cmd->retries);


	sdi_ccon = mrq->cmd->opcode & SAMCOP_SDICMDCON_INDEX;
	sdi_ccon|= SAMCOP_SDICMDCON_SENDERHOST;
	sdi_ccon|= SAMCOP_SDICMDCON_CMDSTART;

	sdi_carg = mrq->cmd->arg;

	//FIXME: Timer value ?!
	sdi_timer= 0xF000;

	sdi_bsize= 0;
	sdi_dcon = 0;
	sdi_imsk = 0;

	//enable interrupts for transmission errors
	sdi_imsk |= SAMCOP_SDIIMSK_RESPONSEND;
	sdi_imsk |= SAMCOP_SDIIMSK_CRCSTATUS;


	host->complete_what = COMPLETION_CMDSENT;

	if (mrq->cmd->flags & MMC_RSP_MASK) {
		host->complete_what = COMPLETION_RSPFIN;

		sdi_ccon |= SAMCOP_SDICMDCON_WAITRSP;
		sdi_imsk |= SAMCOP_SDIIMSK_CMDTIMEOUT;

	} else {
		//We need the CMDSENT-Interrupt only if we want are not waiting
		//for a response
		sdi_imsk |= SAMCOP_SDIIMSK_CMDSENT;
	}

	if(mrq->cmd->flags & MMC_RSP_LONG) {
		sdi_ccon|= SAMCOP_SDICMDCON_LONGRSP;
	}

	if(mrq->cmd->flags & MMC_RSP_CRC) {
		sdi_imsk |= SAMCOP_SDIIMSK_RESPONSECRC;
	}


	if (mrq->data) {
		host->complete_what = COMPLETION_XFERFINISH_RSPFIN;

		sdi_bsize = (1 << mrq->data->blksz_bits);

		sdi_dcon  = (mrq->data->blocks & SAMCOP_SDIDCON_BLKNUM_MASK);
		sdi_dcon |= SAMCOP_SDIDCON_DMAEN;

		sdi_imsk |= SAMCOP_SDIIMSK_FIFOFAIL;
		sdi_imsk |= SAMCOP_SDIIMSK_DATACRC;
		sdi_imsk |= SAMCOP_SDIIMSK_DATATIMEOUT;
		sdi_imsk |= SAMCOP_SDIIMSK_DATAFINISH;
		sdi_imsk |= 0xFFFFFFE0;

		DBG(PFX "request: [DAT] bsize:%u blocks:%u bytes:%u\n",
			sdi_bsize, mrq->data->blocks, mrq->data->blocks * sdi_bsize);

		if(mmc->ios.bus_width == MMC_BUS_WIDTH_4) {
			sdi_dcon |= SAMCOP_SDIDCON_WIDEBUS;
		}

		if(!(mrq->data->flags & MMC_DATA_STREAM)) {
			sdi_dcon |= SAMCOP_SDIDCON_BLOCKMODE;
		}

		if(mrq->data->flags & MMC_DATA_WRITE) {
			sdi_dcon |= SAMCOP_SDIDCON_TXAFTERRESP;
			sdi_dcon |= SAMCOP_SDIDCON_XFER_TXSTART;
		}

		if(mrq->data->flags & MMC_DATA_READ) {
			sdi_dcon |= SAMCOP_SDIDCON_RXAFTERCMD;
			sdi_dcon |= SAMCOP_SDIDCON_XFER_RXSTART;
		}
		
		samcop_sdi_dma_setup(host, mrq->data);
		dma_len = dma_map_sg(&host->sdev->device, mrq->data->sg,
			mrq->data->sg_len, mrq->data->flags & MMC_DATA_READ ?
			DMA_FROM_DEVICE : DMA_TO_DEVICE);

		//start DMA
		samcop_dma_enqueue(host->dma, (void *) host,
			sg_dma_address(&mrq->data->sg[0]),
			sg_dma_len(&mrq->data->sg[0]) );
	}

	host->mrq = mrq;

	init_completion(&host->complete_request);
	init_completion(&host->complete_dma);

	//Clear command and data status registers
	sdi_wrreg(host, SAMCOP_SDICMDSTAT, 0xFFFFFFFF);
	sdi_wrreg(host, SAMCOP_SDIDSTA, 0xFFFFFFFF);

	// Setup SDI controller
	sdi_wrreg(host, SAMCOP_SDIBSIZE, sdi_bsize);
	sdi_wrreg(host, SAMCOP_SDITIMER, sdi_timer);
	sdi_wrreg(host, SAMCOP_SDIIMSK, sdi_imsk);

	// Setup SDI command argument and data control
	sdi_wrreg(host, SAMCOP_SDICMDARG, sdi_carg);
	sdi_wrreg(host, SAMCOP_SDIDCON, sdi_dcon);

	// This initiates transfer
	sdi_wrreg(host, SAMCOP_SDICMDCON, sdi_ccon);

	// Wait for transfer to complete
	wait_for_completion(&host->complete_request);
	DBG("[CMD] request complete.\n");
	if(mrq->data) {
		wait_for_completion(&host->complete_dma);
		DBG("[DAT] DMA complete.\n");
	}
	
	// Wait for DMA to complete
//	if(mrq->data) wait_for_completion(&host->complete_dma);

	//Cleanup controller
	sdi_wrreg(host, SAMCOP_SDICMDARG, 0);
	sdi_wrreg(host, SAMCOP_SDIDCON, 0);
	sdi_wrreg(host, SAMCOP_SDICMDCON, 0);
	sdi_wrreg(host, SAMCOP_SDIIMSK, 0);

	// Read response
	mrq->cmd->resp[0] = sdi_rdreg(host, SAMCOP_SDIRSP0);
	mrq->cmd->resp[1] = sdi_rdreg(host, SAMCOP_SDIRSP1);
	mrq->cmd->resp[2] = sdi_rdreg(host, SAMCOP_SDIRSP2);
	mrq->cmd->resp[3] = sdi_rdreg(host, SAMCOP_SDIRSP3);

	host->mrq = NULL;

	DBG(PFX "request done.\n");

	// If we have no data transfer we are finished here
	if (!mrq->data) goto request_done;

	dma_unmap_sg(&host->sdev->device, mrq->data->sg, dma_len,
			mrq->data->flags & MMC_DATA_READ ? 
			DMA_FROM_DEVICE : DMA_TO_DEVICE);

	// Calulate the amout of bytes transfer, but only if there was
	// no error
	if(mrq->data->error == MMC_ERR_NONE)
		mrq->data->bytes_xfered = (mrq->data->blocks << mrq->data->blksz_bits);
	else
		mrq->data->bytes_xfered = 0;

	// If we had an error while transfering data we flush the
	// DMA channel to clear out any garbage
	if(mrq->data->error != MMC_ERR_NONE) {
		samcop_dma_ctrl(host->dma, SAMCOP_DMAOP_FLUSH);
		DBG(PFX "flushing DMA.\n");		
	}
	// Issue stop command
	if(mrq->data->stop) mmc_wait_for_cmd(mmc, mrq->data->stop, 3);


request_done:

	mrq->done(mrq);
}

static void samcop_sdi_set_ios(struct mmc_host *mmc, struct mmc_ios *ios) {
	struct samcop_sdi_host *host = mmc_priv(mmc);
	u32 sdi_psc, sdi_con;
	int clk_rate;

	//Set power
	sdi_con = sdi_rdreg(host, SAMCOP_SDICON);
	switch(ios->power_mode) {
		case MMC_POWER_ON:
		case MMC_POWER_UP:

			switch(host->asic_type) {

			case SAMCOP_SDI_ASIC_SAMCOP:
#if 0
				/* GPIO_NR_H5400_POWER_SD_N
				   GPIO_NR_H5400_SD_WP_N
				   GPIO_NR_H5400_IRDA_SD */
				SET_H5400_GPIO_N(POWER_SD_N, 1);

				samcop_set_spcr(&h5400_samcop.dev,
					SAMCOP_GPIO_SPCR_SDPUCR, 0);
#endif
				break;

			case SAMCOP_SDI_ASIC_HAMCOP:

				/* XXX Configure SD_DETECT_N and SD_POWER_EN
				       GPIOs here? */
				SET_H2200_GPIO(SD_POWER_EN, 1);

				hamcop_set_spucr(&h2200_hamcop.dev,
					HAMCOP_GPIO_SPUCR_PU_SDCM |
					HAMCOP_GPIO_SPUCR_PU_SDDA,
					0);

				/* XXX Should we only do this if
				   mmc->ios.bus_width == MMC_BUS_WIDTH_4? */
				sdi_con |= HAMCOP_SDICON_CLOCKTYPE_MMC;
				break;
			}

			sdi_con|= SAMCOP_SDICON_FIFORESET;
			break;

		case MMC_POWER_OFF:
		default:
			switch(host->asic_type) {

			case SAMCOP_SDI_ASIC_SAMCOP:
#if 0
				SET_H5400_GPIO_N(POWER_SD_N, 0);
				samcop_set_spcr(&h5400_samcop.dev,
					SAMCOP_GPIO_SPCR_SDPUCR,
					SAMCOP_GPIO_SPCR_SDPUCR);
#endif
				break;

			case SAMCOP_SDI_ASIC_HAMCOP:
				SET_H2200_GPIO(SD_POWER_EN, 0);
				hamcop_set_spucr(&h2200_hamcop.dev,
					HAMCOP_GPIO_SPUCR_PU_SDCM |
					HAMCOP_GPIO_SPUCR_PU_SDDA,
					HAMCOP_GPIO_SPUCR_PU_SDCM |
					HAMCOP_GPIO_SPUCR_PU_SDDA);
				break;
			}

			break;
	}

	//Set clock
	clk_rate = shamcop_gclk(host->parent);
	for(sdi_psc=1;sdi_psc<255;sdi_psc++)
		if( (clk_rate / (sdi_psc+1)) <= ios->clock) break;

	if(sdi_psc > 255) sdi_psc = 255;
	sdi_wrreg(host, SAMCOP_SDIPRE, sdi_psc);

	//Set CLOCK_ENABLE
	if (host->asic_type == SAMCOP_SDI_ASIC_SAMCOP) {
		if(ios->clock) 	sdi_con |= SAMCOP_SDICON_CLOCKTYPE;
		else		sdi_con &=~SAMCOP_SDICON_CLOCKTYPE;
	}

	sdi_wrreg(host, SAMCOP_SDICON, sdi_con);
}

static struct mmc_host_ops samcop_sdi_ops = {
	.request	= samcop_sdi_request,
	.set_ios	= samcop_sdi_set_ios,
};

static u64 samcop_sdi_dmamask = 0xffffffffUL;

static int samcop_sdi_probe(struct device *dev)
{
	struct soc_device *sdev = to_soc_device(dev);
	struct samcop_sdi_host 	*host;
	struct mmc_host 	*mmc;
	int ret;

	mmc = mmc_alloc_host(sizeof(struct samcop_sdi_host), dev);
	if (!mmc) {
		ret = -ENOMEM;
		goto probe_out;
	}

	host = mmc_priv(mmc);

	spin_lock_init( &host->complete_lock );
	host->parent		= dev->parent;
	host->complete_what 	= COMPLETION_NONE;
	host->mmc 		= mmc;
	host->dma		= SAMCOP_SDI_DMA; /* DMA channel # to use */
	host->irq_cd		= H2200_IRQ(SD_DETECT_N); /* XXX _IRQ_SAMCOP_SD_WAKEUP*/
	host->asic_type		= SAMCOP_SDI_ASIC_HAMCOP; /* XXX */

	host->mem = &sdev->resource[0];
	host->base = (void *)sdev->resource[1].start;
	host->irq = (unsigned long)sdev->resource[2].start;
	if(request_irq(host->irq, samcop_sdi_irq, SA_INTERRUPT, DRIVER_NAME,
		       host)) {
		printk(KERN_INFO PFX "failed to request sdi interrupt.\n");
		ret = -ENOENT;
		goto probe_free_host;
	}

#ifdef FIXME
	s3c2410_gpio_cfgpin(S3C2410_GPF2, S3C2410_GPF2_EINT2);
#endif
	set_irq_type(host->irq_cd, IRQT_BOTHEDGE);

	if(request_irq(host->irq_cd, samcop_sdi_irq_cd, 0, DRIVER_NAME " card detect", host)) {
		printk(KERN_WARNING PFX "failed to request card detect interrupt.\n" );
		ret = -ENOENT;
		goto probe_free_irq;
	}

	if(samcop_dma_request(host->dma, &samcop_sdi_dma_client, NULL)) {
		printk(KERN_WARNING PFX "unable to get DMA channel.\n" );
		ret = -EBUSY;
		goto probe_free_irq_cd;
	}

	shamcop_clock_enable(host->parent, SHAMCOP_SD_CLKEN, 1);

	/* HAMCOP's SRAM is controlled by the NAND clock. */
	/* XXX If the NAND driver turns off the clock, we'll have problems. */
	shamcop_clock_enable(host->parent, SHAMCOP_NAND_CLKEN, 1);

	mmc->ops 	= &samcop_sdi_ops;
	mmc->ocr_avail	= MMC_VDD_32_33;
	mmc->caps	= MMC_CAP_4_BIT_DATA;
	mmc->f_min 	= shamcop_gclk(host->parent) / 256;
	mmc->f_max 	= shamcop_gclk(host->parent) / 2;

	/*
	 * SAMCOP has 32KB of SRAM, and HAMCOP 16KB. Remember to save some
	 * space on SAMCOP for the other DMA consumers (USB, etc.).
	 * 32 * 512 = 16K
	 */
	mmc->max_sectors = 32;

	/*
	 * Set the maximum segment size.
	 */
	mmc->max_seg_size = mmc->max_sectors << 9;

	host->sdev = sdev;
	sdev->device.dma_mask = &samcop_sdi_dmamask;
	sdev->device.coherent_dma_mask = 0xffffffffUL;

	dmabounce_register_dev(&sdev->device, 512, 4096);

	printk(KERN_INFO PFX "probe: mapped sdi_base=%p irq=%u irq_cd=%u dma=%u.\n", 
		host->base, host->irq, host->irq_cd, host->dma);

	if((ret = mmc_add_host(mmc))) {
		printk(KERN_INFO PFX "failed to add mmc host.\n");
		goto unreg_dmabounce;
	}

	dev_set_drvdata(dev, mmc);

	printk(KERN_INFO PFX "initialisation done.\n");
	return 0;

 unreg_dmabounce:

	dmabounce_unregister_dev(&sdev->device);

	shamcop_clock_enable(host->parent, SHAMCOP_SD_CLKEN, 0);
	shamcop_clock_enable(host->parent, SHAMCOP_NAND_CLKEN, 0);

	samcop_dma_free(host->dma, &samcop_sdi_dma_client);

 probe_free_irq_cd:
 	free_irq(host->irq_cd, host);

 probe_free_irq:
 	free_irq(host->irq, host);

 probe_free_host:
	mmc_free_host(mmc);
 probe_out:
	return ret;
}

static int samcop_sdi_remove(struct device *dev)
{
	struct soc_device	*sdev = to_soc_device(dev);
	struct mmc_host 	*mmc  = dev_get_drvdata(dev);
	struct samcop_sdi_host 	*host = mmc_priv(mmc);

	mmc_remove_host(mmc);

	dmabounce_unregister_dev(&sdev->device);

	samcop_dma_free(host->dma, &samcop_sdi_dma_client);
	shamcop_clock_enable(host->parent, SHAMCOP_SD_CLKEN, 0);
	shamcop_clock_enable(host->parent, SHAMCOP_NAND_CLKEN, 0);
 	free_irq(host->irq_cd, host);
 	free_irq(host->irq, host);
	mmc_free_host(mmc);

	return 0;
}

/* ------------------- device registration ----------------------- */

static soc_device_id samcop_sdi_soc_device_ids[] = { {IPAQ_SAMCOP_SDI_DEVICE_ID}, {0}};

struct soc_device_driver samcop_sdi_device = {
	.device_ids = samcop_sdi_soc_device_ids,
	.driver = {
		.name = DRIVER_NAME,
		.probe = samcop_sdi_probe,
		.remove = samcop_sdi_remove,
#ifdef CONFIG_PM
//		.suspend = samcop_sdi_suspend,
//		.resume = samcop_sdi_resume,
#endif
        }
};


static int __init samcop_sdi_init(void)
{
	return soc_driver_register (&samcop_sdi_device);
	return 0;
}

static void __exit samcop_sdi_exit(void)
{
	return soc_driver_unregister (&samcop_sdi_device);
}

module_init(samcop_sdi_init);
module_exit(samcop_sdi_exit);

MODULE_DESCRIPTION("SAMCOP SDI Interface driver");
MODULE_AUTHOR("Thomas Kleffel");
MODULE_LICENSE("GPL");
